@@ -4,7 +4,7 @@ use Mojo::Base 'Mojolicious::Plugin';
 use JSON::Validator::OpenAPI;
 use constant DEBUG => $ENV{MOJO_OPENAPI_DEBUG} || 0;
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 my $EXCEPTION = {errors => [{message => 'Internal server error.', path => '/'}], status => 500};
 my $NOT_IMPLEMENTED = {errors => [{message => 'Not implemented.', path => '/'}], status => 501};
@@ -16,10 +16,9 @@ sub register {
   my ($self, $app, $config) = @_;
   my $api_spec = $self->_load_spec($app, $config);
 
-  $app->helper('openapi.input'    => \&_input);
-  $app->helper('openapi.spec'     => sub { shift->stash('openapi.op_spec') });
-  $app->helper('openapi.validate' => sub { $self->_validate(@_) });
-  $app->helper('reply.openapi'    => \&_reply);
+  $app->helper('openapi.invalid_input' => sub { $self->_validate_request(@_) });
+  $app->helper('openapi.spec'          => sub { shift->stash('openapi.op_spec') });
+  $app->helper('reply.openapi'         => sub { $self->_reply(@_) });
   $app->hook(before_render => \&_before_render);
 
   $self->{log_level} = $ENV{MOJO_OPENAPI_LOG_LEVEL} || $config->{log_level} || 'warn';
@@ -45,27 +44,23 @@ sub _add_routes {
 
     for my $http_method (keys %{$paths->{$path}}) {
       next if $http_method =~ $X_RE;
-      my $route_path = $path;
-      my $op_spec    = $paths->{$path}{$http_method};
-      my $name       = $op_spec->{'x-mojo-name'} || $op_spec->{operationId};
-      my $to         = $op_spec->{'x-mojo-to'};
-      my $parameters = $op_spec->{parameters} || [];
-      my %parameters = map { ($_->{name}, $_) } @{$op_spec->{parameters} || []};
+      my $op_spec = $paths->{$path}{$http_method};
+      my $name    = $op_spec->{'x-mojo-name'} || $op_spec->{operationId};
+      my $to      = $op_spec->{'x-mojo-to'};
       my $endpoint;
 
-      $route_path =~ s/{([^}]+)}/{
-        my $name = $1;
-        my $type = $parameters{$name}{'x-mojo-placeholder'} || ':';
-        "($type$name)";
-      }/ge;
+      if ($name and $endpoint = $route->root->find($name)) {
+        $route->add_child($endpoint);
+      }
+      if (!$endpoint) {
+        $endpoint = $route->any(_route_path($path, $op_spec));
+        $endpoint->name($name) if $name;
+      }
 
-      $endpoint = $route->root->find($name) if $name;
-      $endpoint ? $route->add_child($endpoint) : ($endpoint = $route->any($route_path));
       $endpoint->to(ref $to eq 'ARRAY' ? @$to : $to) if $to;
       $endpoint->to($_ => $_->{default})
-        for grep { $_->{in} eq 'path' and exists $_->{default} } @$parameters;
+        for grep { $_->{in} eq 'path' and exists $_->{default} } @{$op_spec->{parameters} || []};
       $endpoint->to({'openapi.op_spec' => $op_spec});
-      $endpoint->name($name) if $name;
       warn "[OpenAPI] Add route $http_method @{[$endpoint->render]}\n" if DEBUG;
     }
   }
@@ -79,7 +74,7 @@ sub _before_render {
   return if !$args->{exception} and grep {/^\w+$/} keys %$args;
   return unless $c->stash('openapi.op_spec');
   my $format = $c->stash('format') || 'json';
-  my $io = $args->{exception} ? $EXCEPTION : $c->stash('openapi.io') || $NOT_IMPLEMENTED;
+  my $io = $args->{exception} ? $EXCEPTION : $NOT_IMPLEMENTED;
   $args->{status} = delete $io->{status};
 
   # TODO: Is $format a good idea? Was thinking someone might want to set
@@ -87,14 +82,6 @@ sub _before_render {
   $args->{$format} = $io;
 }
 
-sub _input {
-  my $c     = shift;
-  my $stash = $c->stash;
-  return $stash->{'openapi.input'} if $stash->{'openapi.input'};
-  return undef if $c->openapi->validate;
-  return $stash->{'openapi.input'};
-}
-
 sub _load_spec {
   my ($self, $app, $config) = @_;
   my $jv       = JSON::Validator->new;
@@ -119,11 +106,13 @@ sub _log {
 }
 
 sub _reply {
-  my ($c, $output, $status) = @_;
+  my ($self, $c, $status, $output) = @_;
   my $format = $c->stash('format') || 'json';
-  $status ||= 200;
-  return $c->render if $c->openapi->validate($output, $status);
-  return $c->render($format => $output, status => $status);
+  return $c->render($format => $output, status => $status)
+    unless my @errors
+    = $self->_validator->validate_response($c, $c->openapi->spec, $status, $output);
+  $self->_log($c, '>>>', \@errors);
+  $c->render(json => {errors => \@errors, status => 500}, status => 500);
 }
 
 sub _reply_spec {
@@ -136,27 +125,26 @@ sub _reply_spec {
   $c->render(json => $spec);
 }
 
-sub _validate {
-  my ($self, $c, $output, $status) = @_;
-  my $op_spec = $c->openapi->spec;
-  my @errors;
+sub _route_path {
+  my ($path, $op_spec) = @_;
+  my %parameters = map { ($_->{name}, $_) } @{$op_spec->{parameters} || []};
+  $path =~ s/{([^}]+)}/{
+    my $pname = $1;
+    my $type = $parameters{$pname}{'x-mojo-placeholder'} || ':';
+    "($type$pname)";
+  }/ge;
+  return $path;
+}
 
-  if (@_ > 2) {
-    $status ||= 200;
-    if (@errors = $self->_validator->validate_response($c, $op_spec, $status, $output)) {
-      $c->stash('openapi.io' => {errors => \@errors, status => 500});
-      $self->_log($c, '>>>', \@errors);
-    }
-  }
-  else {
-    @errors = $self->_validator->validate_request($c, $op_spec, \my %input);
-    if (@errors) {
-      $c->stash('openapi.io' => {errors => \@errors, status => 400});
-      $self->_log($c, '<<<', \@errors);
-    }
-    else {
-      $c->stash('openapi.input' => \%input);
-    }
+sub _validate_request {
+  my ($self, $c, $args) = @_;
+  my $op_spec = $c->openapi->spec;
+  my @errors = $self->_validator->validate_request($c, $op_spec, $c->validation->output);
+
+  if (@errors) {
+    $self->_log($c, '<<<', \@errors);
+    $c->render(json => {errors => \@errors, status => 400}, status => 400)
+      if $args->{auto_render} // 1;
   }
 
   return @errors;
@@ -172,28 +160,37 @@ Mojolicious::Plugin::OpenAPI - OpenAPI / Swagger plugin for Mojolicious
 
 =head1 SYNOPSIS
 
-=head2 Specification
+  use Mojolicious::Lite;
 
-This plugin reads an L<OpenAPI specification|https://openapis.org/specification>
-and generate routes and input/output rules from it. See L<JSON::Validator> for
-L<supported schema formats|JSON::Validator/Supported schema formats>.
+  # Will be moved under "basePath", resulting in "POST /api/echo"
+  post "/echo" => sub {
+    my $c = shift;
+    return if $c->openapi->invalid_input;
+    return $c->reply->openapi(200 => $c->validation->param("body"));
+  }, "echo";
 
+  # Load specification and start web server
+  plugin OpenAPI => {url => "data://main/api.json"};
+  app->start;
+
+  __DATA__
+  @@ api.json
   {
-    "basePath": "/api",
-    "paths": {
-      "/pets": {
-        "get": {
-          "x-mojo-to": "pet#list",
-          "summary": "Finds pets in the system",
-          "parameters": [],
-          "responses": {
+    "swagger" : "2.0",
+    "info" : { "version": "0.8", "title" : "Pets" },
+    "schemes" : [ "http" ],
+    "basePath" : "/api",
+    "paths" : {
+      "/echo" : {
+        "post" : {
+          "x-mojo-name" : "echo",
+          "parameters" : [
+            { "in": "body", "name": "body", "schema": { "type" : "object" } }
+          ],
+          "responses" : {
             "200": {
-              "description": "Pet response",
-              "schema": { "type": "array", "items": { "type": "object" } }
-            },
-            "default": {
-              "description": "Unexpected error",
-              "schema": { "$ref": "http://git.io/vcKD4#" }
+              "description": "Echo response",
+              "schema": { "type": "object" }
             }
           }
         }
@@ -201,83 +198,40 @@ L<supported schema formats|JSON::Validator/Supported schema formats>.
     }
   }
 
-The non-standard part in the spec above is "x-mojo-to". The "x-mojo-to" key can
-either a plain string, object (hash) or an array. The string and hash will be
-passed directly to L<Mojolicious::Routes::Route/to>, while the array ref, will
-be flattened first. Examples:
-
-  "x-mojo-to": "pet#list"
-  $route->to("pet#list");
-
-  "x-mojo-to": {"controller": "pet", "action": "list", "foo": 123}
-  $route->to({controller => "pet", action => "list", foo => 123);
-
-  "x-mojo-to": ["pet#list", {"foo": 123}]
-  $route->to("pet#list", {foo => 123});
-
-The complete HTTP request for getting the "pet list" will be C<GET /api/pets>
-The first part of the path ("/api") comes from C<basePath>, the second part
-comes from the key under C<paths>, and the HTTP method comes from the key under
-C</pets>.
-
-C<parameters> and C<responses> will be used to define rules for
-L<input|/openapi.input> and L<output|reply.openapi>.
-
-=head2 Application
-
-  package Myapp;
-  use Mojolicious;
-
-  sub startup {
-    my $app = shift;
-    $app->plugin("OpenAPI" => {url => $app->home->rel_file("myapi.json")});
-  }
-
-The first thing in your code that you need to do is to load this plugin and the
-L</Specification>. See L</register> for information about what the plugin
-config can be, in addition to "url".
-
-=head2 Controller
-
-  package Myapp::Controller::Pet;
-
-  sub list {
-    my $c = shift;
-
-    # You might want to introspect the specification for the current route
-    my $spec = $c->openapi->spec;
-    unless ($spec->{'x-opening-hour'} == (localtime)[2]) {
-      return $c->reply->openapi([], 498);
-    }
-
-    # $input will be a hash ref if validated and undef on invalid input
-    my $input = $c->openapi->input or return;
-
-    # $output will be validated by the OpenAPI spec before rendered
-    my $output = {pets => [{name => "kit-e-cat"}]};
-    $c->reply->openapi($output, 200);
-  }
-
-The input and output to the action will only be validated if the
-L</openapi.input> and L</reply.openapi> methods are used.
-
-All OpenAPI powered actions will have L<auto-rendering|/reply.openapi> enabled,
-which means that the C<return;> above will render an error document.
-
 =head1 DESCRIPTION
 
+L<Mojolicious::Plugin::OpenAPI> is L<Mojolicious::Plugin> that add routes and
+input/output validation to your L<Mojolicious> application based on a OpenAPI
+(Swagger) specification.
+
+Have a look at the L</SEE ALSO> for references to more documentation, or jump
+right to the L<tutorial|Mojolicious::Plugin::OpenAPI::Guides::Tutorial>.
+
 L<Mojolicious::Plugin::OpenAPI> will replace L<Mojolicious::Plugin::Swagger2>.
 
 This plugin is currently EXPERIMENTAL.
 
 =head1 HELPERS
 
-=head2 openapi.input
+=head2 openapi.invalid_input
 
-  $hash = $c->openapi->input;
+  @errors = $c->openapi->invalid_input;
+  @errors = $c->openapi->invalid_input({auto_render => 0});
 
-Returns the request parameters if they are L<valid|/openapi.validate>, and
-C<undef> on invalid input.
+Used to validate a request. C<@errors> holds a list of
+L<JSON::Validator::Error> objects or empty list on valid input. Setting
+C<auto_render> to a false value will disable the internal auto rendering. This
+is useful if you want to craft a custom resonse.
+
+Validated input parameters will be copied to
+C<Mojolicious::Controller/validation>, which again can be extracted by the
+"name" in the parameters list from the spec. Example:
+
+  # specification:
+  "parameters": [{"in": "body", "name": "whatever", "schema": {"type": "object"}}],
+
+  # controller
+  my $body = $c->validation->param("whatever");
 
 =head2 openapi.spec
 
@@ -295,29 +249,20 @@ Returns the OpenAPI specification for the current route. Example:
     }
   }
 
-=head2 openapi.validate
-
-  # validate request
-  @errors = $c->openapi->validate;
-
-  # validate response
-  @errors = $c->openapi->validate($output, $http_status);
-
-Used to validate input or output data. Request validation is always done by
-L</openapi.input>.
-
 =head2 reply.openapi
 
-  $c->reply->openapi(\%output, $http_status);
+  $c->reply->openapi($status => $output);
 
-Will L<validate|/openapi.validate> C<%output> before passing it on to
-L<Mojolicious::Controller/render>. Note that C<%output> will be passed on using
+Will L<validate|/openapi.validate> C<$output> before passing it on to
+L<Mojolicious::Controller/render>. Note that C<$output> will be passed on using
 the L<format|Mojolicious::Guides::Rendering/Content type> key in stash, which
 defaults to "json". This also goes for L<auto-rendering|/Controller>. Example:
 
   my $format = $c->stash("format") || "json";
   $c->render($format => \%output);
 
+C<$status> is a HTTP status code.
+
 =head1 METHODS
 
 =head2 register
@@ -389,6 +334,16 @@ the terms of the Artistic License version 2.0.
 
 =head1 SEE ALSO
 
-L<Mojolicious::Plugin::Swagger2>.
+=over 2
+
+=item * L<Mojolicious::Plugin::OpenAPI::Guides::Tutorial>
+
+=item * L<http://thorsen.pm/perl/programming/2015/07/05/mojolicious-swagger2.html>.
+
+=item * L<OpenAPI specification|https://openapis.org/specification>
+
+=item * L<Mojolicious::Plugin::Swagger2>.
+
+=back
 
 =cut
